/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.1
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type RestoreState struct {

	//  The path to the save state file to restore the system from.
	SaveStateFilePath string `json:"SaveStateFilePath,omitempty"`

	//  The ID of the template system to clone this new system off of. An empty  string indicates the system should not be cloned from a template.
	TemplateSystemId string `json:"TemplateSystemId,omitempty"`
}
