// Code generated by go generate; DO NOT EDIT.
package generate

import (
	"net/url"

	"github.com/containers/podman/v4/pkg/bindings/internal/util"
)

// Changed returns true if named field has been set
func (o *KubeOptions) Changed(fieldName string) bool {
	return util.Changed(o, fieldName)
}

// ToParams formats struct fields to be passed to API service
func (o *KubeOptions) ToParams() (url.Values, error) {
	return util.ToParams(o)
}

// WithService set field Service to given value
func (o *KubeOptions) WithService(value bool) *KubeOptions {
	o.Service = &value
	return o
}

// GetService returns value of field Service
func (o *KubeOptions) GetService() bool {
	if o.Service == nil {
		var z bool
		return z
	}
	return *o.Service
}
