/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.4
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type CpuGroupConfig struct {
	GroupId         string             `json:"GroupId,omitempty"`
	Affinity        *CpuGroupAffinity  `json:"Affinity,omitempty"`
	GroupProperties []CpuGroupProperty `json:"GroupProperties,omitempty"`
	// Hypervisor CPU group IDs exposed to clients
	HypervisorGroupId uint64 `json:"HypervisorGroupId,omitempty"`
}
