/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.4
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

import "encoding/json"

type ServiceProperties struct {
	// Changed Properties field to []json.RawMessage from []interface{} to avoid having to
	// remarshal sp.Properties[n] and unmarshal into the type(s) we want.
	Properties []json.RawMessage `json:"Properties,omitempty"`
}
