/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.4
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type ContainerCredentialGuardAddInstanceRequest struct {
	Id             string `json:"Id,omitempty"`
	CredentialSpec string `json:"CredentialSpec,omitempty"`
	Transport      string `json:"Transport,omitempty"`
}
