// SPDX-License-Identifier: MIT

syntax = "proto2";

// This one contains statistics about dump/restore process
message dump_stats_entry {
	required uint32			freezing_time		= 1;
	required uint32			frozen_time		= 2;
	required uint32			memdump_time		= 3;
	required uint32			memwrite_time		= 4;

	required uint64			pages_scanned		= 5;
	required uint64			pages_skipped_parent	= 6;
	required uint64			pages_written		= 7;

	optional uint32			irmap_resolve		= 8;

	required uint64			pages_lazy		= 9;
	optional uint64			page_pipes		= 10;
	optional uint64			page_pipe_bufs		= 11;

	optional uint64			shpages_scanned		= 12;
	optional uint64			shpages_skipped_parent	= 13;
	optional uint64			shpages_written		= 14;
}

message restore_stats_entry {
	required uint64			pages_compared		= 1;
	required uint64			pages_skipped_cow	= 2;

	required uint32			forking_time		= 3;
	required uint32			restore_time		= 4;

	optional uint64			pages_restored		= 5;
}

message stats_entry {
	optional dump_stats_entry	dump			= 1;
	optional restore_stats_entry	restore			= 2;
}
